; Make three points with x and y coordinates:
(define p1 (make-object "point" 'x 0  'y 0))
(define p2 (make-object "point" 'x 0  'y 5))
(define p3 (make-object "point" 'x 10 'y 0))

; Add a DistanceTo method to the first point:
(send p1 'AddMember 'DistanceTo
   (function (otherPoint) 
     (sqrt (+ (square (- x (dot otherPoint x))) (square (- y (dot otherPoint y)))))))

(send p1 'DistanceTo p2) ; 5
(send p1 'DistanceTo p3) ; 10
(send p2 'DistanceTo p3)


; List of distances between the points
(list (send p1 'DistanceTo p2) (send p2 'DistanceTo p3) (send p3 'DistanceTo p1))    ; (5 11.180339887498949 10)

; Make three line segment objects in between the points:
(define line-p1-p2 (make-object "line-segment" 'p1 p1 'p2 p2))
(define line-p2-p3 (make-object "line-segment" 'p1 p2 'p2 p3))
(define line-p3-p1 (make-object "line-segment" 'p1 p3 'p2 p1))

; Equip one of these with a Length method:
(send line-p1-p2 'AddMember 'Length (function () (send p1 'DistanceTo p2)))

(map (lambda (ls) (send ls 'Length)) (objects-with-role "line-segment"))   ; (11.180339887498949 5 10)

; Also equip the line-segment objects with a MidPoint method:
(send line-p1-p2 'AddMember 'MidPoint
  (function () 
    (let ((mx (/ (+ (dot p1 x) (dot p2 x)) 2))
          (my (/ (+ (dot p1 y) (dot p2 y)) 2)))
      (make-object "point" 'x mx 'y my))))

; Send the MidPoint messages to all line segments:
(send line-p1-p2 'MidPoint)   ; (object "point" (x . 0) (y . 5/2))
(send line-p2-p3 'MidPoint)   ; (object "point" (x . 5) (y . 5/2))
(send line-p3-p1 'MidPoint)   ; (object "point" (x . 5) (y . 0))

; A a Move method in the point p1, which mutates the x and y coordinates of the point:
(send p1 'AddMember 'Move 
  (function (dx dy)
     (set! x (+ dx x))
     (set! y (+ dy y))))

; Make a triangle with given corners:
(define t1 (make-object "triangle" 'c1 p1 'c2 p2 'c3 p3))

; AUX Scheme definitions:
(define (other-corner-1 corner)
 (cond ((eq? corner 'c1) 'c2)
       ((eq? corner 'c2) 'c3)
       ((eq? corner 'c3) 'c1)
       (else (error "Illegal corner given til other-corner-1"))))

(define (other-corner-2 corner)
 (cond ((eq? corner 'c1) 'c3)
       ((eq? corner 'c2) 'c1)
       ((eq? corner 'c3) 'c2)
       (else (error "Illegal corner given til other-corner-1"))))

(define PI 3.1415962)

(define (radian-to-degree r)
  (/ (* r 360) (* 2 PI)))

(define (degree-to-radians d)
  (/ (* d 2 PI) 360)) 

; We derive a line-segment class object, because we soon need to make a number of line segments internal to triangle calculation methods:
(define line-segment (kappa line-p1-p2))

; All objects of line segment role becomes instances of the line-segment class:
(for-each (lambda (ls) (as-instance-of-class! ls line-segment))
          (objects-with-role "line-segment")) 

; Add an Angle method to t1 which calculates a given angle of the triangle (in degrees).
(send t1 'AddMember 'Angle 
  (function (corner-symbol)
    (let ((adjecent-line-seg-1 (new-object line-segment 'p1 (send this-object 'GetField corner-symbol) 'p2 (send this-object 'GetField (other-corner-1 corner-symbol))))
          (adjecent-line-seg-2 (new-object line-segment 'p1 (send this-object 'GetField corner-symbol) 'p2 (send this-object 'GetField (other-corner-2 corner-symbol))))
          (opposing-line-seg   (new-object line-segment 'p1 (send this-object 'GetField (other-corner-1 corner-symbol)) 'p2 (send this-object 'GetField (other-corner-2 corner-symbol)))))
     (let ((a (send adjecent-line-seg-1 'Length))
           (b (send adjecent-line-seg-2 'Length))
           (c (send opposing-line-seg 'Length)))
       (radian-to-degree (acos (/ (+ (square a) (square b) (- (square c))) (* 2 a b))))))))

; The angles of t1:
(list (send t1 'Angle 'c1) (send t1 'Angle 'c2) (send t1 'Angle 'c3))  ; (89.99989840294607 63.434877213989594 26.56502118895646)

(send t1 'AddMember 'AngleSum 
  (function ()
    (+ (send this-object 'Angle 'c1) (send this-object 'Angle 'c2) (send this-object 'Angle 'c3))))

; Make a new triangle
(define t2 (make-object "triangle" 'c1 p2 'c2 (make-object "point" 'x 0 'y 11) 'c3 (make-object "point" 'x 10 'y 8)))

; The angles of t2:
(list (send t2 'Angle 'c1) (send t2 'Angle 'c2) (send t2 'Angle 'c3))   ; (73.30067301999706 73.30067301999706 33.39845076589797)

; We program a Circumference method for a sample triangle
(send t2 'AddMember 'Circumference
  (function ()
    (+ (send c1 'DistanceTo c2) 
       (send c2 'DistanceTo c3)
       (send c3 'DistanceTo c1))))

; We make yet another triangle
(define t3 (make-object "triangle" 'c1 p1 'c2 (make-object "point" 'x 2 'y 2) 'c3 (make-object "point" 'x 2 'y -2)))

; ... and calculates its circumference by borrowing the method from another object of triangle role.
; Notice that internally, the DistanceTo method of points are borrowed similarly from one of the first points that we made.
(send t3 'Circumference)  ; 9.65685424949238

; We program the Area method, and add to one of the triangles t3 (we know the area is 4):
(send t3 'AddMember 'Area
  (function ()
    (let ((line-seg-1 (new-object line-segment 'p1 (send this-object 'GetField 'c1) 'p2 (send this-object 'GetField 'c2)))
          (line-seg-2 (new-object line-segment 'p1 (send this-object 'GetField 'c1) 'p2 (send this-object 'GetField 'c3)))
          (angle-in-between (degree-to-radians (send this-object 'Angle 'c1))))
      (/ (* (send line-seg-1 'Length) (send line-seg-2 'Length) (sin angle-in-between)) 2))))


(send t3 'Area)  ; 4.000000000000001


; At this point in time we decide to derive the remaining class abstractions, for future programming with points, line-segments, and triangles.
; We have already constructed the line-segment class.
; It is important that we base the classes on objects with all necessary methods. We can use point p1:

(define point (kappa p1))

; Unfortunately, the triangle methods have been sprinkled on three different triangle objects t1, t2, and t3.
; Thefore we need to derive a class from more than one triangle:  Idea  (define triangle (kappa t1 t2 t3))
; We have generalized kappa to accept a number of objects:

(define triangle (kappa t1 t2 t3))

; The following instance of point is structurally equal to t1, because t1 is the first triangle passed to kappa:
(define t4 (new-object triangle))

; Try out the new triangle:
(send t4 'Area)           ; 25.0
(send t4 'Circumference) ; 26.18033988749895
(send t4 'Angle 'c1)     ; 89.99989840294607
